function [mu,Cov_mats] = FKF_forward(A,V,B,W,y,theta)

% executes the forward fast Kalman filter for the state space model
% x_t = Ax_{t-1} + v_t,   v_t ~ N(0,V)
% y_t = B_t*x_t + w_t,    w_t ~ N(0,W_t)
% with threshold theta.
% if theta = 1, then performs the exact Kalman filter algorithm

% Inputs:
% A:  dynamics matrix d x d
% V:  dynamics covariance d x d
% B:  measurement matrices (cell of length T, each matrix has size b_t x d)
% W:  measurement covariance (cell of length T, each matrix b_t x b_t)
% y:  observation vector (cell of length T, each matrix has size b_t x 1)
% theta:    threshold

% Outputs:
% mu:   forward mean vector d x T
% Cov_mats: forward covariance
% if theta < 1, then Cov_mats has size T x 3 with entries C_0, L_t, S_t and
% C_t ~ C_0 - L_t*S_t*L_t'
% if theta = 1, then Cov_mats has size T X 1 with entries C_t

% Written by Eftychios A. Pnevmatikakis, eftychios@stat.columbia.edu
% August 2012

d = size(A,1);
T = length(B);
mu = zeros(d,T);        % forward mean
C0 = V*spdiags(diag(speye(d)-A*A).^-1,0,d,d);    % C_0

if theta < 1  % run fast Kalman filter
    Cov_mats = cell(T,3);
    C0invA = C0\A;

    S = 1;
    L = zeros(d,1);
    for t=1:T   
        AL = A*L;
        Bt = B{t};
        if t > 1                       
            mu_pr = mu(:,t-1);
        else
            mu_pr = zeros(d,1);
        end;
        if isempty(Bt)
            mu(:,t) = A*mu_pr;
            L = A*L;
        else
            PtB = C0*Bt' - AL*S*(Bt*AL)';
            mu(:,t) = A*mu_pr + PtB*((W{t} + Bt*PtB)\(y{t}-Bt*A*mu_pr));
            Phi = C0invA*L;
            Dtinv = (diag(1./diag(S)) - AL'*Phi);
            O = [Phi Bt'];
            Qinv = [Dtinv zeros(size(Dtinv,1),size(Bt,1)); zeros(size(Bt,1),size(Dtinv,1)) W{t}];
            Gt = C0*O*(Qinv + O'*C0*O)^(-0.5);
            [Lh,Sh] = svd(Gt,'econ');   
            Sh=diag(Sh).^2;
            k=find(cumsum(Sh)/sum(Sh)>=theta,1); %grab enough singular vals to cover theta of variance

            L = Lh(:,1:k);
            S = spdiags(Sh(1:k),0,k,k);
        end
        if(~mod(t,100)) fprintf('t=%i; retained singular values = %i; \n',t,k); end;

        Cov_mats{t,1} = C0;
        Cov_mats{t,2} = L;
        Cov_mats{t,3} = S;
    end;
elseif theta == 1
    Cov_mats = cell(T,1);
    for t = 1:T
        Bt = B{t};
        if t == 1
            C_pr = C0;
            mu_pr = zeros(d,1);
        else
            C_pr = Cov_mats{t-1};
            mu_pr = mu(:,t-1);
        end
        Pt = A*C_pr*A' + V;     % predictive covariance
        PtB = Pt*Bt';
        Cov_mats{t} = Pt - PtB*((Bt*PtB + W{t})\PtB');
        mu(:,t) = A*mu_pr + PtB*((Bt*PtB + W{t})\(y{t} - Bt*A*mu_pr));
    end    
else
    error('Threshold value must satisfy 0 < theta <=1');
end
