function [mu,Cov_mats] = LRBT_solver(A,V,B,W,y,theta,varargin)

% mu = LRBT_solver(A,V,B,W,y,theta) executes the Low-Rank Block Thomas
% (LRBT) algorithm for a threshold theta and return the smoothed estimate.
% If theta == 1, then it runs the exact BT algorithm. 

% [mu,Cov_mats] = LRBT_solver(A,V,B,W,y,theta) also provides the smoothed 
% covariance matrices by using a low rank (or the exact) version of the 
% Rybicki Hummer algorithm. If theta < 1, then Cov_mats has size T x 3 
% with entries J_t, L_t, S_t and  C_t ~ J_t - L_t*S_t*L_t'
% if theta = 1, then Cov_mats has size T X 1 with entries C_t

% Inputs:
% A:  dynamics matrix d x d
% V:  dynamics covariance d x d
% B:  measurement matrices (cell of length T, each matrix has size b_t x d)
% W:  measurement covariance (cell of length T, each matrix b_t x b_t)
% y:  observation vector (cell of length T, each matrix has size b_t x 1)
% theta:    threshold


% [mu,Cov_mats] = LRBT_solver(A,V,B,W,y,theta,x0)
% returns the search direction \tilde{H}^{-1}*grad_{x0} used for the 
% iterative application of the LRBT smoother (default x0 = 0 which is used
% in the plain LRBT algorithm

% [mu,Cov_mats] = LRBT_solver(A,V,B,W,y,theta,x0,V0) uses V0 as the 
% covariance of the initial state (default: V0 = V)

% mu = LRBT_solver(A,V,B,W,y,theta,x0,V0,1) treats x0 as a gradient vector
% and returns \tilde{H}^{-1}*x0. This option is used when LRBT is used as a
% preconditioner in the conjugate gradient solver.

% Written by Eftychios A. Pnevmatikakis, eftychios@stat.columbia.edu
% August 2012

d = size(A,1);
T = length(B);
Vinv = V\speye(d);

if nargin < 7 || isempty(varargin{1})
    x0 = zeros(d,T);
else
    x0 = varargin{1};
    if isvector(x0)
        x0 = reshape(x0,d,T);
    end
end
if nargin < 8 || isempty(varargin{2})
    V0 = V;
    V0inv = Vinv;
else
    V0 = varargin{2};
    V0inv = V0\speye(d);
end

if nargin < 9 || isempty(varargin{3}) || ~(varargin{3}==1)
    pcg_flag = 0;
else
    pcg_flag = 1;
end

if nargin < 10 || isempty(varargin{4})
    print_flag = 1;
else
    print_flag = 0;
end

Dm = A'*Vinv*A + Vinv;
Em = A'*Vinv;

if ~pcg_flag || ~any(x0(:))
    grad = grad_compute(x0);%,A,V,B,W,y);
else
    grad = x0(:);
end

Dg = cell(T,1);

Dg{1} = Dm-Vinv+V0inv;
for t = 2:T-1
    Dg{t} = Dm;
end
Dg{T} = Vinv;

if (theta < 1) && (theta > 0)
   low_rank = 1;  % solve with exact Block Thomas Algorithm
elseif theta == 1
   low_rank = 0;
else
   error('Threshold value must satisfy 0 < theta <=1'); 
end


% compute newton direction
q = zeros(d,T);
mu = zeros(d,T);
if low_rank
    L = cell(T,1);
    S = cell(T,1);
    Dt = cell(T,1);  % \tilde{D}_t
    Dt{1} = Dg{1};

    if isempty(B{1})
        L{1} = zeros(d,1);
        S{1} = 1;
    else
        [Lh,Sh] = svd((Dt{1}\B{1}')*(W{1} + B{1}*(Dt{1}\B{1}'))^(-0.5),'econ');
        L{1} = Lh;
        S{1} = spdiags(diag(Sh).^2,0,size(Sh,1),size(Sh,1));
    end
    q(:,1) = -Dt{1}\grad(1:d) + L{1}*S{1}*(L{1}'*grad(1:d));

    for t = 2:T
        Bt = B{t};
        Dt{t} = Dg{t} - Em'*(Dt{t-1}\Em);
        Ot = [Em'*L{t-1} Bt'];
        DtOt = Dt{t}\Ot;
        Qinv = [diag(1./diag(S{t-1})) zeros(size(S{t-1},1),size(Bt,1)); zeros(size(Bt,1),size(S{t-1},1)) W{t}]; 
        Gt = DtOt*(Qinv + Ot'*DtOt)^(-0.5);
        [Lh,Sh] = svd(full(Gt),'econ');
        Sh=diag(Sh).^2;
        k=find(cumsum(Sh)/sum(Sh)>=theta,1); %grab enough singular vals to cover theta of variance
        L{t} = Lh(:,1:k);
        S{t} = spdiags(Sh(1:k),0,k,k);
        if(~mod(t,100)); if(print_flag); fprintf('t=%i; retained singular values = %i; \n',t,k); end; end;
        gEq = grad((t-1)*d+1:t*d) - Em'*q(:,t-1);
        q(:,t) = -Dt{t}\gEq + L{t}*S{t}*(L{t}'*gEq);
    end
    mu(:,T) = q(:,T);
    for t=T-1:-1:1
        mu(:,t) = q(:,t) + (Dt{t}\Em')*mu(:,t+1) - L{t}*S{t}*(L{t}'*Em*mu(:,t+1));
    end
    
    if nargout == 2
        Cov_mats = cell(T,3);
        Lb = cell(T,1);
        Sb = cell(T,1);
        Dtb = cell(T,1);
        Dtb{T} = Dg{T};
        if isempty(B{T})
            Lb{T} = zeros(d,1);
            Sb{T} = 1;
        else
            [Lh,Sh] = svd((Dtb{T}\B{T}')*(W{T} + B{T}*(Dt{T}\B{T}'))^(-0.5),'econ');
            Lb{T} = Lh;
            Sb{T} = spdiags(diag(Sh).^2,0,size(Sh,1),size(Sh,1));            
        end
        for t = T-1:-1:2
            Bt = B{t};
            Dtb{t} = Dg{t} - Em*(Dtb{t+1}\Em');
            Obt = [Em*Lb{t+1} Bt'];
            DtOt = Dtb{t+1}\Obt;
            Qbinv = [diag(1./diag(Sb{t+1})) zeros(size(Sb{t+1},1),size(Bt,1)); zeros(size(Bt,1),size(Sb{t+1},1)) W{t}];
            Gbt = DtOt*(Qbinv + Obt'*DtOt)^(-0.5);
            [Lbh,Sbh] = svd(full(Gbt),'econ');
            Sbh = diag(Sbh).^2;
            k = find(cumsum(Sbh)/sum(Sbh)>=theta,1);
            Lb{t} = Lbh(:,1:k);
            Sb{t} = spdiags(Sbh(1:k),0,k,k);
            if(~mod(t,100)); fprintf('t=%i; retained singular values = %i; \n',t,k); end;
        end
        for t = 1:T-1
            DtEm = Dt{t}\Em;
            DbEm = Dtb{t+1}\Em';
            J_mat = speye(d) - DtEm*DbEm;
            P_mat = [L{t}*S{t}, DtEm*Lb{t+1}*Sb{t+1}-L{t}*S{t}*(L{t}'*Em*Lb{t+1}*Sb{t+1})];
            R_mat = [L{t}'*Em*DbEm; Lb{t+1}'*Em']';
            Dtinv = Dt{t}\speye(d);
            Cov_mats{t,1} = J_mat\Dtinv;
            M1 = J_mat\P_mat;
            M2 = (speye(size(P_mat,2)) + R_mat'*M1)\(R_mat'*(J_mat\speye(d)));
            N2 = S{t}*L{t}';
            
            %Cov_mats{t,2} = [M1, J_mat\L{t}];
            %Cov_mats{t,3} = [M2*Dtinv-(M2*L{t})*N2; N2];
            
            A1 = [M1, J_mat\L{t}];
            A2 = [M2*Dtinv-(M2*L{t})*N2; N2];
            [Q,R] = qr(A1,0); 
            [P,Z] = qr(A2',0); 
            [U,G2,~] = svd(full(R*Z'));
            dG2 = diag(G2);
            k = find(cumsum(dG2)/sum(dG2)>=1-1e-8,1);  % enforce stability
            Cov_mats{t,2} = Q*U(:,1:k); 
            Cov_mats{t,3} = spdiags(dG2(1:k),0,k,k);             
        end
            Cov_mats{T,1} = Dt{T}\speye(d);
            Cov_mats{T,2} = L{T};
            %Cov_mats{T,3} = S{T}*L{T}';
            Cov_mats{T,3} = S{T};
    end
else  % run exact BT algorithm
    Mtinv = cell(1,T);
    Mtinv{1} = inv(Dg{1} + B{1}'*(W{1}\B{1}));
    q(:,1) = -Mtinv{1}*grad(1:d); 
    
    for t = 2:T
        Bt = B{t};
        Mtinv{t} = inv(Dg{t}+ Bt'*(W{t}\Bt) - Em'*Mtinv{t-1}*Em);
        q(:,t) = -Mtinv{t}*(grad((t-1)*d+1:t*d) - Em'*q(:,t-1));        
    end
    mu(:,T) = q(:,T);
    for t=T-1:-1:1
        mu(:,t) = q(:,t) + Mtinv{t}*Em*mu(:,t+1);
    end
    if nargout == 2 % estimate covariance as well
        Cov_mats = cell(T,1);
        Ntinv = cell(T,1);
        Ntinv{T} = inv(Dg{T} + B{T}'*(W{T}\B{T}));
        for t = T-1:-1:2
            Ntinv{t} = inv(Dg{t} + B{t}'*(W{t}\B{t}) - Em*Ntinv{t+1}*Em');
        end
        for t = 1:T-1
            Cov_mats{t} = (speye(d) - Mtinv{t}*Em*Ntinv{t+1}*Em')\Mtinv{t};
        end
        Cov_mats{T} = Mtinv{T};
    end
end

if(pcg_flag); mu = mu(:); end;

%%%%% gradient computing function

function grad = grad_compute(x) %,A,V,B,W,y)

    % computes the gradient of the negative log-likelihood function at point x
    
    grad = zeros(d*T,1);

    for t = 1:T
        grad((t-1)*d + (1:d)) = - B{t}'*(W{t}\(y{t}-B{t}*x(:,t)));
    end

    if any(x(:))
        grad(1:d) = grad(1:d) + V0inv*x(:,1) - Em*(x(:,2)-A*x(:,1));
        for t = 2:T-1
            grad((t-1)*d + (1:d)) = grad((t-1)*d + (1:d)) - Em*(x(:,t+1)-A*x(:,t)) + Vinv*(x(:,t)-A*x(:,t-1));
        end
        grad((T-1)*d + (1:d)) = grad((T-1)*d + (1:d)) + Vinv*(x(:,T)-A*x(:,T-1));
    end
end

end