% Runs a demo similar to the receptive field experiment and applies all the
% solvers discussed in the paper. (fast kalman filter, exact kalman
% filter, low rank block Thomas, exact block Thomas, iterative LRBT,
% conjugate gradient with a LRBT preconditioner.

% written by Eftychios A. Pnevmatikakis, Columbia University
% August 2012.

clear; clc; 
close all;
d=2^8; 
d2=2*d;                 % the actual size of the receptive field grid (must be even)

dx=1/d2; xx=dx:dx:1;    % spatial domain
rf_width=.2;            % width of receptive field
T = 500;               % number of time steps
tau = 50;               % time constant
rho=exp(-1/tau);        % discrete time constant
A=rho*speye(d2);        % dynamics matrix
theta=1-1e-4;           % threshold on how many singular values to keep

%specify power spectrum
spec = zeros(d2,1);
spec(1) = 1;                   % DC part of power spectrum
spec(2:d+1) = (1:d).^-2;       % real part of power spectrum
spec(d+2:end) = spec(2:d);     % imag part (same as real part, to enforce stationarity)
spec = spec/sum(spec/d2);      % normalize by total power per spatial bin
V = .1*spdiags(spec,0,d2,d2);  % dynamics covariance  (not proportional to identity)   
C0 = (speye(d2) - A*A')\V;     % prior covariance

%RR maps the elements of q into the fft of the target signal
RR = spalloc(d2,d2,d2*4);       % zeros(d2);
RR(1,1) = 1;                    % zero frequency
RR(d+1,d+1) = 1; 
RR(d2+2+(0:d2+1:(d-2)*(d2+1))) = 1;
RR(d2*(d+1)+2+(0:d2+1:(d-2)*(d2+1))) = 1i;
RR(d2*2+(0:d2-1:(d-2)*(d2-1))) = 1;
RR(d2*(d+2)+(0:d2-1:(d-2)*(d2-1))) = -1i;
C0_var_rf = trace(RR*C0*RR')/d2;  % prior variance in spatial domain

here = zeros(T,1);        % represents the location of mean of the Gaussian bump of the receptive field
n = cell(T,1);            % spike train
rf = zeros(d2,T);         % receptive field
sn = 1e-1;                % observation variance (assumed white for simplicity)
%bt = ones(T,1);           % # of observations for each timestep (fixed)
bt = randi([0,3],T,1);    % random number of observations per time step (variable)

disp('making data...')      
B = cell(T,1);            % measurements matrix
W = cell(T,1);
for t=1:T
    %update the RF 
    here(t,:)=.5*(1-0.5*sin(2*2*pi*t/T));                   % mean of the Gaussian bump of the RF..
    rf(:,t)=5*exp( -.5*((xx-here(t,1)).^2)/rf_width^2);     % value of the RF at time t
    Bss=randn(bt(t),d2)/sqrt(d2);                           % random white-noise stimulus case
    B{t} = (real(RR'*fft(Bss'))/sqrt(d2))';                 % transform stimulus into frequency domain   
    W{t} = sn*speye(bt(t));
    n{t} = Bss*rf(:,t) + sqrt(sn)*randn(bt(t),1);          % observation model
end;
fprintf('done. \n')


%% Run the algorithms

fprintf('Now run the various inference algorithms. \nStart from the fast Kalman filter algorithm with threshold %f. \n',theta)
tt1 = tic;
[mu_FKF,Cov_FKF] = FKF_forward(A,V,B,W,n,theta);
mu_RF_FKF=real(ifft(RR*mu_FKF))*sqrt(d2);       % transform back to spatial domain
t_FKF = toc(tt1);
var_RF_FKF = zeros(d2,T);
for t = 1:T
    sqrt_DUBB=sqrt(Cov_FKF{t,3})*real(ifft(RR*Cov_FKF{t,2}))'*sqrt(d2);
    var_RF_FKF(:,t)=sum(sqrt_DUBB.^2,1)'; %diag_C0_rf+diag(BB'*U*D*U'*BB)';
end
fprintf('done after %2.2f seconds.\n',t_FKF)

tt1 = tic;
disp('... Now running the exact forward Kalman filter for comparison')    
[mu_KF_ex,Cov_KF_ex] = FKF_forward(A,V,B,W,n,1);
mu_RF_KF_ex=real(ifft(RR*mu_KF_ex))*sqrt(d2); 
t_KF_ex = toc(tt1);
fprintf('done after %2.2f seconds.\n',t_KF_ex)


tt1 = tic;
disp('... Now run the LRBT smoother with the same threshold')
[mu_LRBT,Cov_LRBT] = LRBT_solver(A,V,B,W,n,theta,[],C0);
mu_RF_LRBT = real(ifft(RR*mu_LRBT))*sqrt(d2); 
t_LRBT = toc(tt1);
var_RF_LRBT = zeros(d2,T);
for t = 1:T
    sqrt_DUBB=sqrt(Cov_LRBT{t,3})*real(ifft(RR*Cov_LRBT{t,2}))'*sqrt(d2);
    %var_RF_LRBT(:,t)=trace(RR*Cov_LRBT{t,1}*RR')/d2-sum(sqrt_DUBB.^2,1)'; %diag_C0_rf+diag(BB'*U*D*U'*BB)';
    var_RF_LRBT(:,t)= sum(sqrt_DUBB.^2,1)';
end
fprintf('done after %2.2f seconds.\n',t_LRBT)

disp('... Now run the exact BT smoother');
tt1 = tic;
[mu_BT_ex,Cov_BT_ex] = LRBT_solver(A,V,B,W,n,1,[],C0);
mu_RF_BT_ex = real(ifft(RR*mu_BT_ex))*sqrt(d2); 
t_BT_ex = toc(tt1);
fprintf('done after %2.2f seconds.\n',t_BT_ex)

tt1 = tic;
disp('... Now run the iterative LRBT smoother with the same threshold')
mu_iter = iterative_LRBT_solver(A,V,B,W,n,theta,[],V);
mu_RF_iter = real(ifft(RR*mu_iter))*sqrt(d2); 
t_iter = toc(tt1);
fprintf('done after %2.2f seconds.\n',t_iter)


tt1 = tic;
disp('... Now run the pcg solver with the LRBT preconditioner')
[mu_pcg,iter,resvec] = pcg_LRBT_solver(A,V,B,W,n,theta);
mu_RF_pcg = real(ifft(RR*mu_pcg))*sqrt(d2);
t_pcg = toc(tt1);
fprintf('done after %2.2f seconds.\n',t_pcg)

mV = max([rf(:);mu_RF_FKF(:);mu_RF_KF_ex(:);mu_RF_LRBT(:);mu_RF_BT_ex(:);mu_RF_iter(:);mu_RF_pcg(:)]);
nV = min([rf(:);mu_RF_FKF(:);mu_RF_KF_ex(:);mu_RF_LRBT(:);mu_RF_BT_ex(:);mu_RF_iter(:);mu_RF_pcg(:)]);

%% Do some plotting
disp('Now do some plotting')

figure;  % plot filtered and smoothed means
    subplot(711); imagesc(rf,[nV,mV]); colorbar; title('Original');
    subplot(712); imagesc(mu_RF_FKF,[nV,mV]); colorbar; title(sprintf('Fast Kalman Filter. Time: %2.2f sec',t_FKF));
    subplot(713); imagesc(mu_RF_KF_ex,[nV,mV]); colorbar; title(sprintf('Exact Kalman Filter. Time: %2.2f sec',t_KF_ex));
    subplot(714); imagesc(mu_RF_LRBT,[nV,mV]); colorbar; title(sprintf('LRBT Smoother. Time: %2.2f sec',t_LRBT));
    subplot(715); imagesc(mu_RF_BT_ex,[nV,mV]); colorbar; title(sprintf('Exact BT Smoother. Time: %2.2f sec',t_BT_ex));
    subplot(716); imagesc(mu_RF_iter,[nV,mV]); colorbar; title(sprintf('Iterative LRBT Smoother. Time: %2.2f sec',t_iter));
    subplot(717); imagesc(mu_RF_pcg,[nV,mV]); colorbar; title(sprintf('PCG solver with LRBT preconditioner. Time: %2.2f sec',t_pcg));
 
t = ceil(rand*T); % pick a random timestep
    
figure;  % plot filtered and smoothed covariances
    subplot(4,2,[1,2]); imagesc(var_RF_FKF); colorbar; title('Forward Variance Reduction');
    subplot(4,2,[3,4]); imagesc(var_RF_LRBT); colorbar; title('Smoothed Variance Reduction');
%     subplot(4,2,5); imagesc(C0 - Cov_FKF{t,2}*Cov_FKF{t,3}*Cov_FKF{t,2}'); axis square; colorbar; title(sprintf('Approximate Forward Covariance at t = %i',t));
%     subplot(4,2,6); imagesc(Cov_KF_ex{t}); axis square; colorbar; title(sprintf('Exact Forward Covariance at t = %i',t));
%     subplot(4,2,7); imagesc(Cov_LRBT{t,1} - Cov_LRBT{t,2}*Cov_LRBT{t,3}*Cov_LRBT{t,2}'); axis square; colorbar; title(sprintf('Approximate Smoothed Covariance at t = %i',t));
%     subplot(4,2,8); imagesc(Cov_BT_ex{t}); axis square; colorbar; title(sprintf('Exact Smoothed Covariance at t = %i',t));    
    subplot(4,2,[5,7]); imagesc(Cov_KF_ex{t}\(C0 - Cov_FKF{t,2}*Cov_FKF{t,3}*Cov_FKF{t,2}')); axis square; colorbar; title(sprintf('C_{ex}^{-1}*C_{app} at t = %i (Forward Covariance)',t));
    %subplot(4,2,6); imagesc(Cov_KF_ex{t}); axis square; colorbar; title(sprintf('Exact Forward Covariance at t = %i',t));
    %subplot(4,2,7); imagesc(Cov_LRBT{t,1} - Cov_LRBT{t,2}*Cov_LRBT{t,3}*Cov_LRBT{t,2}'); axis square; colorbar; title(sprintf('Approximate Smoothed Covariance at t = %i',t));
    subplot(4,2,[6,8]); imagesc(Cov_BT_ex{t}\(Cov_LRBT{t,1} - Cov_LRBT{t,2}*Cov_LRBT{t,3}*Cov_LRBT{t,2}')); axis square; colorbar; title(sprintf('C_{ex}^{-1}*C_{app} at t = %i (Smoothed Covariance)',t)); 

    
%% Check the appoximation error for various thresholds

if (1)
    disp('... Now checking the FKF and LRBT approximation error for various thresholds') 
    c = -2:-1:-6;  % theta = 1 - 10^c(i)
    lc = length(c);
    err_KF = zeros(lc,1);
    err_BT = zeros(lc,1);
    nKF = norm(mu_KF_ex,'fro');
    nBT = norm(mu_BT_ex,'fro');
    for i = 1:lc
        FKF_sol = FKF_forward(A,V,B,W,n,1 - 10^c(i)); 
        err_KF(i) = norm(FKF_sol-mu_KF_ex,'fro')/nKF;
        LRBT_sol = LRBT_solver(A,V,B,W,n,1 - 10^c(i),[],C0); 
        err_BT(i) = norm(LRBT_sol-mu_BT_ex,'fro')/nBT;
    end
    figure;
        subplot(121); loglog(10.^c,err_KF); xlabel('1-\theta'); ylabel('Relative error'); 
            title('Relative error bewteen approximate and exact solutions (Fast Kalman)');
            set(gca,'XDir','reverse'); grid on;
        subplot(122); loglog(10.^c,err_BT); xlabel('1-\theta'); ylabel('Relative error'); 
            title('Relative error bewteen approximate and exact solutions (LRBT)');
            set(gca,'XDir','reverse'); grid on;
end    