function [mu,iter,resvec] = pcg_LRBT_solver(A,V,B,W,y,theta,varargin)

% Conjugate gradient using a Low-Rank Block-Thomas preconditioner. 
% mu = pcg_LRBT_solver(A,V,B,W,y,theta), returns the smoothed mean using 
% conjugate gradients with a LRBT preconditioner.

% Inputs:
% A:  dynamics matrix d x d
% V:  dynamics covariance d x d
% B:  measurement matrices (cell of length T, each matrix has size b_t x d)
% W:  measurement covariance (cell of length T, each matrix b_t x b_t)
% y:  observation vector (cell of length T, each matrix has size b_t x 1)
% theta: threshold

% mu = pcg_LRBT_solver(A,V,B,W,y,theta,x0,V0) uses x0 as the initial
% point (default x0 = zero vector) and V0 as the covariance of the initial 
% state (default: V0 = V).

% x0: initial point (default: zero vector)
% V0: covariance of the initial state (default: V0 = V)

% [mu,iter,resvec] = pcg_LRBT_solver(A,V,B,W,y,theta,varargin) also returns
% the number of iterations needed for convergence and the vector of residuals.

% Written by Eftychios A. Pnevmatikakis, eftychios@stat.columbia.edu
% August 2012

d = size(A,1);
T = length(B);
Vinv = V\speye(d);

if nargin < 7 || isempty(varargin{1})
    x0 = zeros(d,T);
    x = x0;
    grad = grad_compute(x);
    grad0 = grad;
else
    x = varargin{1};%x0;
    grad = grad_compute(x);
    grad0 = grad_compute(zeros(d,T));
end
if nargin < 8
    V0 = V;
    V0inv = Vinv;
else
    V0 = varargin{2};
    V0inv = V0\speye(d);
end

Dm = A'*Vinv*A + Vinv;
Em = A'*Vinv;

Dg = cell(T,1);

Dg{1} = Dm-Vinv+V0inv;
for t = 2:T-1
    Dg{t} = Dm;
end
Dg{T} = Vinv;

Hx = @(x) hessian_vector(x);
LRBT =@(x) -LRBT_solver(A,V,B,W,y,theta,x,V0,1,0);
MaxIter = 100;
[mu,~,~,iter,resvec] = pcg(Hx,-grad,[],MaxIter,LRBT);
mu = reshape(mu,d,T);


%%%%% Hessian x vector computing function

function Hx = hessian_vector(x)
    
    % computes the product of the Hessian H with a vector x in O(dT) time
    % without explicitly constructing the Hessian
    
    x = reshape(x,d,T);
    % computes the product H*x 
    Ex = -Em*[x(:,2:end) zeros(d,1)] - Em'*[zeros(d,1) x(:,1:end-1)];
    Hx = Ex(:);
    for t = 1:T
        Hx((t-1)*d+(1:d)) = Hx((t-1)*d+(1:d)) + Dg{t}*x(:,t) + B{t}'*(W{t}\(B{t}*x(:,t)));
    end
end
   
%%%%% gradient computing function

function grad = grad_compute(x)

    % computes the gradient of the negative log-likelihood function at point x
    
    grad = zeros(d*T,1);

    for t = 1:T
        grad((t-1)*d + (1:d)) = - B{t}'*(W{t}\(y{t}-B{t}*x(:,t)));
    end

    if any(x(:))
        grad(1:d) = grad(1:d) + V0inv*x(:,1) - Em*(x(:,2)-A*x(:,1));
        for t = 2:T-1
            grad((t-1)*d + (1:d)) = grad((t-1)*d + (1:d)) - Em*(x(:,t+1)-A*x(:,t)) + Vinv*(x(:,t)-A*x(:,t-1));
        end
        grad((T-1)*d + (1:d)) = grad((T-1)*d + (1:d)) + Vinv*(x(:,T)-A*x(:,T-1));
    end
end

end